/*
 * linux/kernel/chr_drv/keyboard.c
 *
 * Keyboard driver for Linux v0.96 using Latin-1.
 *
 * Written for linux by Johan Myreen as a translation from
 * the assembly version by Linus (with diacriticals added)
 *
 * Loadable keymaps by Risto Kankkunen, May 1993
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file README.legal in the main directory of this archive
 * for more details.
 */

/*
 * general 680x0 support by Hamish Macdonald
 */

#ifndef KBD_DEFFLAGS
#ifdef CONFIG_KBD_META
#define KBD_DEFFLAGS ((1 << VC_NUMLOCK) | (1 << VC_REPEAT) | (1 << VC_META))
#else
#define KBD_DEFFLAGS ((1 << VC_NUMLOCK) | (1 << VC_REPEAT))
#endif
#endif

#define SHIFT_KEYS	((1 << KG_LSHIFT) | (1 << KG_RSHIFT))
#define CTRL_KEYS	((1 << KG_LCTRL)  | (1 << KG_RCTRL))
#define ALT_KEYS	((1 << KG_LALT)   | (1 << KG_RALT))
#define ALTGR_KEYS	((1 << KG_LALTGR) | (1 << KG_RALTGR))

#include <linux/sched.h>
#include <linux/tty.h>
#include <linux/mm.h>
#include <linux/keyboard.h>
#include <linux/config.h>
#include <linux/signal.h>

#ifndef KBD_DEFFLAGS
#ifdef CONFIG_KBD_META
#define KBD_DEFFLAGS ((1 << VC_NUMLOCK) | (1 << VC_REPEAT) | (1 << VC_META))
#else
#define KBD_DEFFLAGS ((1 << VC_NUMLOCK) | (1 << VC_REPEAT))
#endif
#endif

extern void do_keyboard_interrupt(void);
extern void change_console(unsigned int new_console);
extern void scrollback(int);
extern void scrollfront(int);

unsigned long kbd_flags = 0;
unsigned long kbd_dead_keys = 0;
unsigned long kbd_prev_dead_keys = 0;

static int want_console = -1;
static int last_console = 0;		   /* last used VC */
static char rep = 0;			   /* flag telling character repeat */
struct kbd_struct kbd_table[NR_CONSOLES];
struct kbd_struct *kbd = kbd_table;
struct tty_struct *tty = NULL;

typedef unsigned short u_word;
typedef void (*k_hand)(unsigned char value, char up_flag);

static void do_self(unsigned char value, char up_flag);
static void do_fn(unsigned char value, char up_flag);
static void do_spec(unsigned char value, char up_flag);
static void do_pad(unsigned char value, char up_flag);
static void do_dead(unsigned char value, char up_flag);
static void do_cons(unsigned char value, char up_flag);
static void do_cur(unsigned char value, char up_flag);
static void do_shift(unsigned char value, char up_flag);

static k_hand key_handler[] = {
	do_self, do_fn, do_spec, do_pad, do_dead, do_cons, do_cur, do_shift
};

/* maximum values each key_handler can handle */
const u_char max_vals[] = {
	255, 25, 13, 16, 4, 255, 3, 255
};

const int NR_TYPES = (sizeof(max_vals) / sizeof(u_char));

static int shift_state = 0;
static int diacr = -1;
static int npadch = 0;
fptr key_table[];

static unsigned int handle_diacr(unsigned int);

void process_scancode (int scancode)
{
	static unsigned char prev_scancode = 0;

	kbd_prev_dead_keys |= kbd_dead_keys;
	if (!kbd_dead_keys)
		kbd_prev_dead_keys = 0;
	kbd_dead_keys = 0;
	tty = TTY_TABLE(0);
	kbd = kbd_table + fg_console;
	if (vc_kbd_flag (kbd,VC_RAW)) {
		kbd_flags = 0;
		put_queue (scancode);
		goto end_kbd_intr;
	}
	/*
	 *  Repeat a key only if the input buffers are empty or the
	 *  characters get echoed locally. This makes key repeat usable
	 *  with slow applications and under heavy loads.
	 */
	rep = scancode == prev_scancode;
	prev_scancode = scancode;
	if (!rep || 
	    (vc_kbd_flag(kbd,VC_REPEAT) && tty &&
	     (L_ECHO(tty) || (EMPTY(&tty->secondary) && EMPTY(&tty->read_q))))) {
		u_word key_code;
		char break_flag = scancode > 0x7f;

		scancode &= 0x7f;
		key_code = key_map[shift_state][scancode];
		(*key_handler[key_code >> 8])(key_code & 0xff, break_flag);
	     }
end_kbd_intr:
	do_keyboard_interrupt();
}

void put_queue(int ch)
{
    struct tty_queue *qp;

    wake_up(&keypress_wait);
    if (!tty)
	return;
    qp = &tty->read_q;

    if (LEFT(qp)) {
	    qp->buf[qp->head] = ch;
	    INC(qp->head);
	    wake_up_interruptible(&qp->proc_list);
    }
}

static void puts_queue(char *cp)
{
    struct tty_queue *qp;
    char ch;

    wake_up(&keypress_wait);
    if (!tty)
	return;
    qp = &tty->read_q;

    while ((ch = *(cp++)) != 0) {
	    if (LEFT(qp)) {
		    qp->buf[qp->head] = ch;
		    INC(qp->head);
	    }
    }
    wake_up_interruptible(&qp->proc_list);
}

static void applkey(int key, char mode)
{
    char buf[] = { 0x1b, 'O', 0x00, 0x00 };

    buf[1] = (mode ? 'O' : '[');
    buf[2] = key;
    puts_queue(buf);
}

static void enter(void)
{
	put_queue(13);
	if (vc_kbd_flag(kbd,VC_CRLF))
		put_queue(10);
}

static void caps_toggle(void)
{
	if (rep)
		return;
	chg_kbd_flag(KG_CAPSLOCK);
	chg_vc_kbd_flag(kbd,VC_CAPSLOCK);
}

static void show_ptregs(void)
{
    return;
}

static void hold(void)
{
	/* kludge, should have a control key table */
	if (kbd_flags & CTRL_KEYS) {
		show_state();
		return;
	}

	if (rep || !tty)
		return;
	if (vc_kbd_flag(kbd, VC_SCROLLOCK))
		/* pressing scroll lock 2nd time sends ^Q, ChN */
		put_queue(START_CHAR(tty));
	else
		/* pressing scroll lock 1st time sends ^S, ChN */
		put_queue(STOP_CHAR(tty));
	chg_vc_kbd_flag(kbd,VC_SCROLLOCK);
}

static void num(void)
{
#if 0
	if (kbd_flags & CTRL_KEYS) {
		/* pause key pressed, sends E1 1D 45, ChN */
		chg_vc_kbd_flag(kbd,VC_PAUSE);
		return;
	}
#endif
	if (vc_kbd_flag(kbd,VC_APPLIC)) {
		applkey('P', 1);
		return;
	}
	if (!rep)	/* no autorepeat for numlock, ChN */
		chg_vc_kbd_flag(kbd,VC_NUMLOCK);
}

static void lastcons(void)
{
	/* pressing alt-printscreen switches to the last used console, ChN */
	want_console = last_console;
}

static void send_intr(void)
{
	if (tty)
		put_queue(INTR_CHAR(tty));
}

static void do_spec(unsigned char value, char up_flag)
{
	typedef void (*fnp)(void);
	fnp fn_table[] = {
		NULL,		enter,		show_ptregs,	show_mem,
		show_state,	send_intr,	lastcons, 	caps_toggle,
		num,		hold
	};

	if (value >= sizeof(fn_table)/sizeof(fnp))
		return;
	if (up_flag)
		return;
	if (!fn_table[value])
		return;
	fn_table[value]();
}
  
static void do_self(unsigned char value, char up_flag)
{
	if (up_flag)
		return;		/* no action, if this is a key release */

	value = handle_diacr(value);

	/* kludge... */
	if (vc_kbd_flag(kbd,VC_CAPSLOCK))
		if ((value >= 'a' && value <= 'z')
		    || (value >= 224 && value <= 254)) {
			value -= 32;
		}
	/* kludge... */
	if (kbd_flags & CTRL_KEYS) {
		if (value >= 64 && value < 127)
			value &= 0x1f;
		else if (value == ' ' || value == '2')
			value = 0;
		else if (value >= '3' && value < '8')
			value -= 24;
		else if (value == '?' || value == '8' || value == '/')
			value = 127;
		else if (value == '-' || value == '_')
			value = 0x1f;
		else
			return;
	}

	if (kbd_flags & ALT_KEYS)
		if (vc_kbd_flag(kbd,VC_META)) {
			put_queue('\033');
			put_queue(value);
		} else
			put_queue(value|0x80);
	else
		put_queue(value);
}

static unsigned char ret_diacr[] =
	{'`', '\'', '^', '~', '"' };		/* Must not end with 0 */

/* If a dead key pressed twice, output a character corresponding it,	*/
/* otherwise just remember the dead key.				*/

static void do_dead(unsigned char value, char up_flag)
{
	if (up_flag)
		return;

	if (diacr == value) {	/* pressed twice */
		diacr = -1;
		put_queue(ret_diacr[value]);
		return;
	}
	diacr = value;
}

/* If no pending dead key, return the character unchanged. Otherwise,	*/
/* if space if pressed, return a character corresponding the pending	*/
/* dead key, otherwise try to combine the two.				*/

unsigned int handle_diacr(unsigned int ch)
{
	static unsigned char accent_table[5][64] = {
	" \300BCD\310FGH\314JKLMN\322PQRST\331VWXYZ[\\]^_"
	"`\340bcd\350fgh\354jklmn\362pqrst\371vwxyz{|}~",   /* accent grave */

	" \301BCD\311FGH\315JKLMN\323PQRST\332VWX\335Z[\\]^_"
	"`\341bcd\351fgh\355jklmn\363pqrst\372vwxyz{|}~",   /* accent acute */

	" \302BCD\312FGH\316JKLMN\324PQRST\333VWXYZ[\\]^_"
	"`\342bcd\352fgh\356jklmn\364pqrst\373vwxyz{|}~",   /* circumflex */

	" \303BCDEFGHIJKLMN\325PQRSTUVWXYZ[\\]^_"
	"`\343bcdefghijklm\361\365pqrstuvwxyz{|}~",	    /* tilde */

	" \304BCD\313FGH\316JKLMN\326PQRST\334VWXYZ[\\]^_"
	"`\344bcd\353fgh\357jklmn\366pqrst\374vwx\377z{|}~" /* dieresis */
	};
	int d = diacr;

	if (diacr == -1)
		return ch;

	diacr = -1;
	if (ch == ' ')
		return ret_diacr[d];

	if (ch < 64 || ch > 122)
		return ch;

	return accent_table[d][ch - 64];
}

static void do_cons(unsigned char value, char up_flag)
{
	if (up_flag)
		return;
	want_console = value;
}

static char *func_table[] = {
	"\033[[A",  "\033[[B",  "\033[[C",  "\033[[D",  "\033[[E", /* F1 -F5  */
	"\033[17~", "\033[18~", "\033[19~", "\033[20~", "\033[21~",/* F6 -F10 */
	"\033[23~", "\033[24~", "\033[25~", "\033[26~", "\033[28~",/* F11-F15 */
	"\033[29~", "\033[31~", "\033[32~", "\033[33~", "\033[34~",/* F16-F20 */
	"\033[1~",	/* Find */
	"\033[2~",	/* Insert */
	"\033[3~",	/* Remove */
	"\033[4~",	/* Select */
	"\033[5~",	/* Prev page */
	"\033[6~"	/* Next page */
};

static void do_fn(unsigned char value, char up_flag)
{
	if (up_flag)
		return;
	if (kbd_flags & SHIFT_KEYS) {
		if (value == KVAL(K_PGDN)) {
			scrollfront(0);
			return;
		}
		if (value == KVAL(K_PGUP)) {
			scrollback(0);
			return;
		}
	}
	if (kbd_flags & ALT_KEYS) {
		want_console = value;
		return;
	}
	puts_queue(func_table[value]);
}

static void do_pad(unsigned char value, char up_flag)
{
	static char *pad_chars = "0123456789+-*/\015,.";
	static char *app_map = "pqrstuvwxylSRQMnn";

	if (up_flag)
		return;		/* no action, if this is a key release */

	if ((value == KVAL(K_PCOMMA) || value == KVAL(K_PDOT)) &&
	    (kbd_flags & CTRL_KEYS) && (kbd_flags & (ALT_KEYS | ALTGR_KEYS))) {
#if 0
		ctrl_alt_del();
#endif
		return;
	}

	if ((kbd_flags & ALT_KEYS) && value <= 9) {	/* Alt-numpad */
		npadch = (npadch * 10 + value) % 1000;
		return;
	}

	/* kludge... shift forces cursor/number keys */
	if (vc_kbd_flag(kbd,VC_APPLIC) && shift_state != 1) {
		applkey(app_map[value], 1);
		return;
	}

	if (!vc_kbd_flag(kbd,VC_NUMLOCK))
		switch (value) {
			case KVAL(K_PCOMMA):
			case KVAL(K_PDOT):
				do_fn(KVAL(K_REMOVE), 0);
				return;
			case KVAL(K_P0):
				do_fn(KVAL(K_INSERT), 0);
				return;
			case KVAL(K_P1):
				do_fn(KVAL(K_SELECT), 0);
				return;
			case KVAL(K_P2):
				do_cur(KVAL(K_DOWN), 0);
				return;
			case KVAL(K_P3):
				do_fn(KVAL(K_PGDN), 0);
				return;
			case KVAL(K_P4):
				do_cur(KVAL(K_LEFT), 0);
				return;
			case KVAL(K_P6):
				do_cur(KVAL(K_RIGHT), 0);
				return;
			case KVAL(K_P7):
				do_fn(KVAL(K_FIND), 0);
				return;
			case KVAL(K_P8):
				do_cur(KVAL(K_UP), 0);
				return;
			case KVAL(K_P9):
				do_fn(KVAL(K_PGUP), 0);
				return;
			case KVAL(K_P5):
				applkey('G', vc_kbd_flag(kbd,VC_APPLIC));
				return;
		}

	put_queue(pad_chars[value]);
	if (value == KVAL(K_PENTER) && vc_kbd_flag(kbd,VC_CRLF))
		put_queue(10);
}

static void do_cur(unsigned char value, char up_flag)
{
	static char *cur_chars = "BDCA";
	if (up_flag)
		return;

	applkey(cur_chars[value], vc_kbd_flag(kbd,VC_CKMODE));
}

static void do_shift(unsigned char value, char up_flag)
{
	shift_state = 0;
	if (up_flag)
		clr_kbd_flag(value);
	else
		set_kbd_flag(value);
	if (kbd_flags & SHIFT_KEYS)
		shift_state = 1;
	if (kbd_flags & ALTGR_KEYS)
		shift_state = 2;
	/* cludge */
	if (up_flag && value == KG_LALT && npadch != 0) {
		put_queue(npadch);
		npadch=0;
	}
}

#define C(x)		((KT_CONS<<8)|x)

u_word key_map[NR_KEYMAPS][NR_KEYS] = {
	{ /* unshifted keys */
	K_HOLE,  27,      '1',     '2',     '3',     '4',     '5',     '6',
	'7',     '8',     '9',     '0',     '-',     '=',     127,       9,
	'q',     'w',     'e',     'r',     't',     'y',     'u',     'i',
	'o',     'p',     '[',     ']', K_ENTER, K_LCTRL,     'a',     's',
	'd',     'f',     'g',     'h',     'j',     'k',     'l',     ';',
	'\'',    '`',  K_LSHIFT,  '\\',     'z',     'x',     'c',     'v',
	'b',     'n',     'm',     ',',     '.',     '/',     K_RSHIFT, K_PSTAR,
	K_ALT,   ' ',     K_CAPS,  K_F1,    K_F2,    K_F3,    K_F4,    K_F5,
	K_F6,    K_F7,    K_F8,    K_F9,    K_F10,   K_NUM,   K_HOLD,  K_P7,
	K_P8,    K_P9,    K_PMINUS, K_P4,   K_P5,    K_P6,    K_PPLUS, K_P1,
	K_P2,    K_P3,    K_P0,    K_PDOT,  K_CONS,  K_HOLE,  8,       K_F11,
	K_F12,   K_HOLE,  K_HOLE,  K_HOLE,  K_HOLE,  K_HOLE,  K_HOLE,  K_HOLE,
	K_PENTER, K_RCTRL, K_PSLASH,    28, K_ALTGR, K_BREAK,  K_FIND,    K_UP,
	K_PGUP,   K_LEFT, K_RIGHT, K_SELECT, K_DOWN,  K_PGDN, K_INSERT, K_REMOVE },
	{ /* shifted keys */
	 K_HOLE,      27,     '!',     '@',     '#',     '$',     '%',     '^',
	    '&',     '*',     '(',     ')',     '_',     '+',     127,       9,
	    'Q',     'W',     'E',     'R',     'T',     'Y',     'U',     'I',
	    'O',     'P',     '{',     '}', K_ENTER,  K_LCTRL,    'A',     'S',
	    'D',     'F',     'G',     'H',     'J',     'K',     'L',     ':',
	    '"',     '~', K_LSHIFT,    '|',     'Z',     'X',     'C',     'V',
	    'B',     'N',     'M',     '<',     '>',     '?', K_RSHIFT, K_PSTAR,
	  K_ALT,      32,  K_CAPS,   K_F11,   K_F12,   K_F13,   K_F14,   K_F15,
	  K_F16,   K_F17,   K_F18,   K_F19,   K_F20,   K_NUM, K_SH_MEM,   K_P7,
	   K_P8,    K_P9, K_PMINUS,   K_P4,    K_P5,    K_P6, K_PPLUS,    K_P1,
	   K_P2,    K_P3,    K_P0,  K_PDOT,  K_CONS,  K_HOLE,       8,   K_F11,
	  K_F12,  K_HOLE,  K_HOLE,  K_HOLE,  K_HOLE,  K_HOLE,  K_HOLE,  K_HOLE,
	K_PENTER, K_RCTRL, K_PSLASH, K_HOLE, K_ALTGR, K_BREAK, K_FIND,  K_UP,
	K_PGUP,   K_LEFT, K_RIGHT, K_SELECT, K_DOWN, K_PGDN, K_INSERT, K_REMOVE },
	{ /* alted keys */
	 K_HOLE,  K_HOLE,  K_HOLE,     '@',  K_HOLE,     '$',  K_HOLE,  K_HOLE,
	    '{',     '[',     ']',     '}',    '\\',  K_HOLE,  K_HOLE,  K_HOLE,
	 K_HOLE,  K_HOLE,  K_HOLE,  K_HOLE,  K_HOLE,  K_HOLE,  K_HOLE,  K_HOLE,
	 K_HOLE,  K_HOLE,  K_HOLE,     '~', K_ENTER, K_LCTRL,  K_HOLE,  K_HOLE,
	 K_HOLE,  K_HOLE,  K_HOLE,  K_HOLE,  K_HOLE,  K_HOLE,  K_HOLE,  K_HOLE,
	 K_HOLE,  K_HOLE, K_LSHIFT, K_HOLE,  K_HOLE,  K_HOLE,  K_HOLE,  K_HOLE,
	 K_HOLE,  K_HOLE,  K_HOLE,  K_HOLE,  K_HOLE,  K_HOLE, K_RSHIFT, K_HOLE,
	  K_ALT,  K_HOLE,  K_CAPS,   C(12),   C(13),   C(14),   C(15),   C(16),
	  C(17),   C(18),   C(19),   C(20),   C(21),   K_NUM, K_SH_REGS,  K_P7,
	   K_P8,    K_P9, K_PMINUS,   K_P4,    K_P5,    K_P6, K_PPLUS,    K_P1,
	   K_P2,    K_P3,    K_P0,  K_PDOT,  K_CONS,  K_HOLE,       8,   C(22),
	  C(23),  K_HOLE,  K_HOLE,  K_HOLE,  K_HOLE,  K_HOLE,  K_HOLE,  K_HOLE,
	K_PENTER, K_RCTRL, K_PSLASH, K_HOLE, K_ALTGR, K_BREAK, K_FIND,    K_UP,
	K_PGUP,   K_LEFT, K_RIGHT, K_SELECT, K_DOWN, K_PGDN, K_INSERT, K_REMOVE }};

unsigned long kbd_init(unsigned long kmem_start)
{
    int i;
    struct kbd_struct * kbd;

    kbd = kbd_table + 0;
    for (i = 0 ; i < NR_CONSOLES ; i++,kbd++) {
	kbd->flags = KBD_DEFFLAGS;
	kbd->default_flags = KBD_DEFFLAGS;
    }

    return mach_keyb_init (kmem_start);
}
